#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2020 University of Liège
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


## Trim analysis
# Adrien Crovato
#
# Find the angle of attack to match a specified lift coefficient

import math
import dart.utils as dU
import tbox.utils as tU
import fwk
from fwk.coloring import ccolors
from ..core import init_dart

class Trim:
    """Utility to perform the trim analysis of a given lifting configuration
    Find the angle of attack to reach a desired lift coefficient

    Attributes
    ----------
    tms : fwk.Timers object
        dictionary of timers
    dim : int
        problem dimensions
    format : str
        output format type
    slice : array of float
        arrays of y-coordinates to perform slices (for 3D only)
    tag : int
        index of physical tag to slice (see gmsh)
    mach : float
        freestream Mach number
    clTgt : float
        target lift coefficient
    alpha : float
        initial angle of attack
    dCl : float
        initial (estimated) slope of lift curve
    """
    def __init__(self, p):
        """Setup and configure

        Parameters
        ----------
        p : dict
            dictionary of parameters to configure DART
        """
        # basic init
        self.tms = fwk.Timers()
        self.tms["total"].start()
        _dart = init_dart(p, scenario='aerodynamic', task='analysis')
        self.dim, self.msh, self.wrt, self.pbl, self.bnd, self.sol = (_dart.get(key) for key in ['dim', 'msh', 'wrt', 'pbl', 'bnd', 'sol'])
        # save some parameters for later use
        self.format = p['Format']
        try:
            self.slice = p['Slice']
            self.tag = p['TagId']
        except:
            self.slice = None
            self.tag = None
        self.mach = p['M_inf']
        self.clTgt = p['CL']
        self.alpha = p['AoA']*math.pi/180
        self.dCl = p['dCL']

    def run(self):
        """Perform the trim analysis
        """
        # initialize loop
        it = 0
        while True:
            # define current angle of attack
            print(ccolors.ANSI_BLUE + 'Setting AoA to', self.alpha*180/math.pi, ccolors.ANSI_RESET)
            # update problem and reset ICs to improve robustness in transonic cases
            self.pbl.update(self.alpha)
            self.sol.reset()
            # run the solver
            self.tms["solver"].start()
            status = self.sol.run()
            if status >= 2:
                raise RuntimeError('Flow solver diverged!\n')
            self.tms["solver"].stop()
            # update slope
            if it != 0:
                self.dCl = (self.sol.Cl - Cl_) / (self.alpha - alpha_)
            # break or store values and update AoA
            if abs(self.clTgt - self.sol.Cl) < 0.005 or math.isnan(self.sol.Cl):
                break
            else:
                Cl_ = self.sol.Cl
                alpha_ = self.alpha
                dAlpha = (self.clTgt - self.sol.Cl) / self.dCl
                self.alpha += dAlpha
                it += 1

        # save results
        self.sol.save(self.wrt)
        # extract Cp
        if self.dim == 2:
            Cp = dU.extract(self.bnd.groups[0].tag.elems, self.sol.Cp)
            tU.write(Cp, f'Cp_{self.msh.name}_airfoil.dat', '%1.5e', ', ', 'x, y, z, Cp', '')
        elif self.dim == 3 and self.format == 'vtk' and self.slice:
            dU.write_slices(self.msh.name, self.slice, self.tag)

    def disp(self):
        """Display the results
        """
        # display results
        print(ccolors.ANSI_BLUE + 'Trim analysis' + ccolors.ANSI_RESET)
        print("       M    alpha      dCl       Cl       Cd       Cm")
        print("{0:8.2f} {1:8.1f} {2:8.4f} {3:8.4f} {4:8.4f} {5:8.4f}".format(self.mach, self.alpha*180/math.pi, self.dCl, self.sol.Cl, self.sol.Cd, self.sol.Cm))
        print('\n')
        # display timers
        self.tms["total"].stop()
        print(ccolors.ANSI_BLUE + 'CPU statistics' + ccolors.ANSI_RESET)
        print(self.tms)
