/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "wBody.h"
#include "wFace.h"
#include "wResults.h"
#include "wGroup.h"
#include "wTag.h"
#include "wElement.h"
#include "wNode.h"
#include <unordered_set>
using namespace tbox;
using namespace dart;

Body::Body(std::shared_ptr<MshData> _msh, std::vector<int> const &nos) : Groups(_msh, nos), Cl(0), Cd(0), Cs(0), Cm(0)
{
    // Sanity check
    if (groups.size() != 2)
    {
        std::stringstream err;
        err << "dart::Body should be built with exactly 2 physical groups but " << groups.size() << " were given!\n";
        throw std::runtime_error(err.str());
    }
    create();
}

Body::Body(std::shared_ptr<MshData> _msh, std::vector<std::string> const &names) : Groups(_msh, names), Cl(0), Cd(0), Cs(0), Cm(0)
{
    // Sanity check
    if (groups.size() != 2)
    {
        std::stringstream err;
        err << "dart::Body should be built with exactly 2 physical groups but " << groups.size() << " were given!\n";
        throw std::runtime_error(err.str());
    }
    create();
}

/**
 * @brief Create data structure
 */
void Body::create()
{
    // Get the nodes of Body
    for (auto e : groups[0]->tag->elems)
        for (auto n : e->nodes)
            nodes.push_back(n);
    std::sort(nodes.begin(), nodes.end());
    auto iterator = std::unique(nodes.begin(), nodes.end());
    nodes.resize(std::distance(nodes.begin(), iterator));
    // Create a map between the body (surface) nodes and their local index
    for (size_t i = 0; i < nodes.size(); ++i)
        nMap[nodes[i]] = i;

    // Create unordered set of surface faces
    std::unordered_set<Face *, EquFace, EquFace> bndFaces;
    for (auto e : groups[0]->tag->elems)
    {
        Face *f = new Face(e->nodes);
        f->el0 = e;
        bndFaces.insert(f);
    }
    // Find volume faces that are surface faces (valid for TRI3 and TETRA4 only)
    for (auto e : groups[1]->tag->elems)
    {
        size_t idx = 0;
        for (size_t i = 0; i < e->nodes.size(); ++i)
        {
            // build volume face
            std::vector<Node *> fN(e->nodes.size() - 1);
            for (size_t j = 0; j < e->nodes.size() - 1; ++j)
                fN[j] = e->nodes[(idx + j) % e->nodes.size()];
            Face f(fN);
            // check if face is a surface face
            // for some meshes, 1 tetra4 can have 2 faces that are surface tri3!
            auto it = bndFaces.find(&f);
            if (it != bndFaces.end())
                (*it)->el1 = e;
            idx++;
        }
    }
    // Create a map between the body (surface) and volume elements
    for (auto f : bndFaces)
    {
        if (f->el1 == NULL)
        {
            std::stringstream err;
            err << "dart::Body: no volume element from tag " << *groups[1]->tag << "could be associated to surface element " << *(f->el0) << "!\n";
            throw std::runtime_error(err.str());
        }
        svMap[f->el0] = f->el1;
        delete f;
    }

    // Resize loads vector
    nLoads.resize(nodes.size(), Eigen::Vector3d::Zero());
}

void Body::write(std::ostream &out) const
{
    out << *groups[0]->tag << " is a Body immersed in " << *groups[1]->tag << std::endl;
}
