/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "wWakeElement.h"
#include "wNode.h"
using namespace tbox;
using namespace dart;

WakeElement::WakeElement(size_t _no, Element *&_surUpE, Element *&_surLwE,
                         Element *&_volUpE, Element *&_volLwE,
                         std::vector<std::pair<size_t, Node *>> &_wkN)
    : wObject(), no(_no), surUpE(_surUpE), surLwE(_surLwE),
      volUpE(_volUpE), volLwE(_volLwE), wkN(_wkN)
{
    // Sanity checks
    if (surUpE->type() != surLwE->type())
        throw std::runtime_error("WakeElement: upper and lower surface element are not the same type!\n");
    if (surUpE->type() == ElType::LINE2)
        nDim = 2;
    else if (surUpE->type() == ElType::TRI3)
        nDim = 3;
    else
    {
        std::stringstream err;
        err << "WakeElement only implemented for surface elements of type "
            << ElType::LINE2 << " or " << ElType::TRI3
            << " (" << surUpE->type() << " was given)!\n";
        throw std::runtime_error(err.str());
    }
    if (volUpE->type() == ElType::TRI3 || volUpE->type() == ElType::TETRA4)
        nColUp = volUpE->nodes.size();
    else
    {
        std::stringstream err;
        err << "WakeElement only implemented for volume elements of type "
            << ElType::TRI3 << " or " << ElType::TETRA4
            << " (" << volUpE->type() << " was given)!\n";
        throw std::runtime_error(err.str());
    }
    if (volLwE->type() == ElType::TRI3 || volLwE->type() == ElType::TETRA4)
        nColLw = volLwE->nodes.size();
    else
    {
        std::stringstream err;
        err << "WakeElement only implemented for volume elements of type "
            << ElType::TRI3 << " or " << ElType::TETRA4
            << " (" << volLwE->type() << " was given)!\n";
        throw std::runtime_error(err.str());
    }
    // Map the nodes
    mapNodes();
    nRow = wkN.size();
}

/**
 * @brief Map the wake nodes to their local volume node indices
 */
void WakeElement::mapNodes()
{
    // Find local volume index of contributing surface nodes
    for (auto p : wkN)
    {
        for (size_t i = 0; i < volUpE->nodes.size(); ++i)
        {
            if (p.second == volUpE->nodes[i])
            {
                vsMap[p.first] = i;
                break;
            }
        }
    }
    // Check that all contributing surface nodes have been mapped
    if (vsMap.size() != wkN.size())
    {
        std::stringstream err;
        err << "dart::WakeElement: could not map some contributing nodes from surface element " << *surUpE << " to " << *volUpE << " (number of mapped nodes: " << vsMap.size() << ")!\n";
        throw std::runtime_error(err.str());
    }
}

void WakeElement::write(std::ostream &out) const
{
    out << "WakeElement #" << no
        << "\n\ton surface elements (" << surUpE->type() << "): " << surUpE->no << " and " << surLwE->no
        << "\n\ton volume elements: (" << volUpE->type() << "): " << volUpE->no << " and " << volLwE->no << std::endl;
}